hr{margin-top:1rem;margin-bottom:1rem;border:0;border-top:1px solid rgba(0,0,0,.1)}.dm-table__cell{align-items:center;box-sizing:border-box;display:flex;flex:150 0 auto;max-width:-webkit-min-content;max-width:-moz-min-content;max-width:min-content;min-width:-webkit-min-content;min-width:-moz-min-content;min-width:min-content;overflow:hidden;padding:5px;position:relative;width:-webkit-fit-content;width:-moz-fit-content;width:fit-content;word-break:break-word}.row_activate:hover,.row_activate_selected{background-color:#0000ff26}.lsf-space_direction_horizontal{align-items:center;grid-auto-flow:column}.lsf-space_size_small{grid-gap:8px}.lsf-space{grid-gap:16px;display:grid}.lsf-userpic{align-items:center;background:#f5f5f5;border-radius:13.5px;box-shadow:inset 0 0 0 1px rgba(0,0,0,.1);display:flex;height:28px;justify-content:center;overflow:hidden;position:relative;-webkit-user-select:none;-moz-user-select:none;user-select:none;width:28px}.lsf-userpic img{font-size:12px;height:100%;line-height:22px;-o-object-fit:cover;object-fit:cover;opacity:0;position:absolute;width:100%}.lsf-annotations-list__create,.lsf-annotations-list__entity{align-items:center;cursor:pointer;display:flex;height:48px;padding:0 16px}.lsf-userpic__username{display:block;font-family:Roboto;font-size:12px;font-weight:700;line-height:22px;opacity:.4}.lsf-space_direction_vertical{grid-auto-flow:row;grid-auto-rows:-webkit-max-content;grid-auto-rows:max-content;justify-content:center}.lsf-space_collapsed,.lsf-space_size_none{grid-gap:0}.lsf-annotations-list__user{display:flex;font-size:14px;line-height:16px;max-width:165px}.lsf-annotations-list__name{flex:1;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.lsf-annotations-list__entity-id{flex:none;font-size:11px;opacity:.5;padding-left:5px}.lsf-annotations-list__created{color:rgba(0,0,0,.4);font-size:11px;line-height:13px}