"use strict";(self["webpackChunkantgo_web"]=self["webpackChunkantgo_web"]||[]).push([[854],{6854:function(t,a,e){e.r(a),e.d(a,{default:function(){return l}});var n=function(){var t=this,a=t._self._c;return a("div",[a("div",{staticClass:"mt-3"},[a("b-pagination",{staticStyle:{display:"flex","flex-direction":"row","justify-content":"center"},attrs:{pills:"","total-rows":t.page_num,"per-page":t.perPage},nativeOn:{click:function(a){return t.next_list()}},model:{value:t.page_index,callback:function(a){t.page_index=a},expression:"page_index"}}),a("div",{staticStyle:{display:"flex","flex-direction":"row","justify-content":"center"}},[a("b-badge",{staticStyle:{"margin-right":"10px"},attrs:{variant:"success"}},[t._v("等待处理 "+t._s(t.waiting))]),a("b-badge",{attrs:{variant:"danger"}},[t._v("完成处理 "+t._s(t.finished))])],1),a("b-container",t._l(t.result_list,(function(e,n){return a("b-row",{key:n,staticStyle:{"margin-bottom":"5px"}},t._l(e["data"],(function(e,n){return a("b-col",{key:n,staticStyle:{display:"flex","justify-content":"center"}},[a("div",{staticClass:"dm-table__cell",staticStyle:{"margin-top":"7px"}},[a("div",{staticStyle:{width:"200px",display:"flex",height:"100%","align-items":"center"}},["IMAGE"==e["type"]?a("div",[a("img",{attrs:{src:e["data"],width:e["width"],height:e["height"]}})]):a("div",[t._v(" "+t._s(e["data"])+" ")])])])])})),1)})),1)],1)])},i=[],s={name:"Predict",data(){return{num_in_page:1,page_num:1,page_index:1,dataset_name:"",sample_num_in_dataset:0,current_dataset_name:"",result_num:0,result_list:[],tags:[],waiting:0,finished:0,perPage:1}},mounted:function(){var t=this;t.axios.get("/antgo/api/predict/entry/").then((function(a){t.num_in_page=a.data.content["num_in_page"],t.page_num=a.data.content["page_num"],t.page_index=a.data.content["page_index"]+1,t.dataset_name=a.data.content["dataset_name"],t.sample_num_in_dataset=a.data.content["sample_num_in_dataset"],t.current_dataset_name=a.data.content["current_dataset_name"],t.result_num=a.data.content["result_num"],t.result_list=a.data.content["result_list"],t.tags=a.data.content["tags"],t.waiting=a.data.content["waiting"],t.finished=a.data.content["finished"]})).catch((function(t){console.log(t)}))},methods:{next_list:function(){var t=this;t.axios.get("/antgo/api/predict/page/",{params:{page_index:t.page_index-1,dataset_name:t.dataset_name[0]}}).then((function(a){t.num_in_page=a.data.content["num_in_page"],t.page_num=a.data.content["page_num"],t.result_num=a.data.content["result_num"],t.result_list=a.data.content["result_list"],t.tags=a.data.content["tags"],t.waiting=a.data.content["waiting"],t.finished=a.data.content["finished"]})).catch((function(t){console.log(t)}))}}},c=s,d=e(1001),_=(0,d.Z)(c,n,i,!1,null,"441c0481",null),l=_.exports}}]);
//# sourceMappingURL=854.d3c0e54f.js.map