import os

highlight = [
    {'color': '#00BFFF','name': 'DeepSkyBlue'},
    {'color': '#00F5FF', 'name': 'Turquoise1'},
    {'color': '#7FFFD4', 'name': 'Aquamarine1'},
    {'color': '#54FF9F', 'name': 'SeaGreen1'},
    {'color': '#00FF00', 'name': 'Green1'},
    {'color': '#FFF68F', 'name': 'Khaki1'},
    {'color': '#FFFFE0', 'name': 'LightYellow1'},
    {'color': '#FFFF00', 'name': 'Yellow1'},
    {'color': '#FFD700', 'name': 'Gold1'},
    {'color': '#FFC1C1', 'name': 'RosyBrown1'},
    {'color': '#FF6A6A', 'name': 'IndianRed1'},
    {'color': '#FF8247', 'name': 'Sienna1'},
    {'color': '#FFD39B', 'name': 'Burlywood1'},
    {'color': '#FFA54F', 'name': 'Tan1'},
    {'color': '#FF3030', 'name': 'Firebrick1'},
    {'color': '#FF8C69', 'name': 'Salmon1'},
    {'color': '#FFA500', 'name': 'Orange1'},
    {'color': '#FF0000', 'name': 'Red1'},
    {'color': '#FF1493', 'name': 'DeepPink1'},
    {'color': '#FF00FF', 'name': 'Magenta1'},
    {'color': '#BF3EFF', 'name': 'DarkOrchid1'},
    {'color': '#9B30FF', 'name': 'Purple1'},
    {'color': '#00FF7F', 'name': 'SpringGreen'},
    {'color': '#FFFFE0', 'name': 'LightYellow'},
    {'color': '#FFDEAD', 'name': 'NavajoWhite1'},
    {'color': '#FFF0F5', 'name': 'LavenderBlush1'},
    {'color': '#836FFF', 'name': 'SlateBlue1'},
    {'color': '#63B8FF', 'name': 'SteelBlue1'},
    {'color': '#00BFFF', 'name': 'DeepSkyBlue1'},
    {'color': '#CAE1FF', 'name': 'LightSteelBlue1'},
    {'color': '#BFEFFF', 'name': 'LightBlue1'},
    {'color': '#E0FFFF', 'name': 'LightCyan1'}
]

dark=[
    {'color': '#0000FF','name': 'Blue'},
    {'color': '#00868B', 'name': 'Turquoise4'},
    {'color': '#458B74', 'name': 'Aquamarine4'},
    {'color': '#2E8B57', 'name': 'SeaGreen4'},
    {'color': '#008B00', 'name': 'Green4'},
    {'color': '#8B864E', 'name': 'Khaki4'},
    {'color': '#8B8B7A', 'name': 'LightYellow4'},
    {'color': '#8B8B00', 'name': 'Yellow4'},
    {'color': '#CDAD00', 'name': 'Gold3'},
    {'color': '#CD9B9B', 'name': 'RosyBrown3'},
    {'color': '#8B3A3A', 'name': 'IndianRed4'},
    {'color': '#8B4726', 'name': 'Sienna4'},
    {'color': '#CDAA7D', 'name': 'Burlywood3'},
    {'color': '#CD853F', 'name': 'Tan3'},
    {'color': '#8B1A1A', 'name': 'Firebrick4'},
    {'color': '#8B4C39', 'name': 'Salmon4'},
    {'color': '#8B5A00', 'name': 'Orange4'},
    {'color': '#8B0000', 'name': 'Red4'},
    {'color': '#8B0A50', 'name': 'DeepPink4'},
    {'color': '#8B008B', 'name': 'Magenta4'},
    {'color': '#68228B', 'name': 'DarkOrchid4'},
    {'color': '#551A8B', 'name': 'Purple4'},
    {'color': '#00FF00', 'name': 'Green'},
    {'color': '#FFFF00', 'name': 'Yellow'},
    {'color': '#CDB38B', 'name': 'NavajoWhite3'},
    {'color': '#CDC1C5', 'name': 'LavenderBlush3'},
    {'color': '#6959CD', 'name': 'SlateBlue3'},
    {'color': '#4F94CD', 'name': 'SteelBlue3'},
    {'color': '#009ACD', 'name': 'DeepSkyBlue3'},
    {'color': '#A2B5CD', 'name': 'LightSteelBlue3'},
    {'color': '#9AC0CD', 'name': 'LightBlue3'},
    {'color': '#B4CDCD', 'name': 'LightCyan3'}
]