.dm-table__cell[data-v-441c0481]{align-items:center;box-sizing:border-box;display:flex;flex:150 0 auto;max-width:-webkit-min-content;max-width:-moz-min-content;max-width:min-content;min-width:-webkit-min-content;min-width:-moz-min-content;min-width:min-content;overflow:hidden;padding:5px;position:relative;width:-webkit-fit-content;width:-moz-fit-content;width:fit-content;word-break:break-word}