# -*- coding: UTF-8 -*-
# @Time    : 2019/1/23 3:13 PM
# @File    : sampling_def.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

"""Abstract class for sampling methods.

Provides interface to sampling methods that allow same signature
for select_batch.  Each subclass implements select_batch_ with the desired
signature for readability.
"""

import abc
import numpy as np

class SamplingMethod(object):
  __metaclass__ = abc.ABCMeta

  def flatten_X(self, X):
    shape = X.shape
    flat_X = X
    if len(shape) > 2:
      flat_X = np.reshape(X, (shape[0], np.product(shape[1:])))
    return flat_X


  @abc.abstractmethod
  def select_batch_(self):
    return

  def select_batch(self, **kwargs):
    return self.select_batch_(**kwargs)

  def to_dict(self):
    return None