# -*- coding: UTF-8 -*-
# @Time    : 2022/5/1 15:03
# @File    : data.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.framework.torch2paddle.mapper.io import *