.dm-table__cell[data-v-500d50b3]{align-items:center;box-sizing:border-box;display:flex;flex:150 0 auto;max-width:-webkit-min-content;max-width:-moz-min-content;max-width:min-content;min-width:-webkit-min-content;min-width:-moz-min-content;min-width:min-content;overflow:hidden;padding:10px;position:relative;width:-webkit-fit-content;width:-moz-fit-content;width:fit-content;word-break:break-word}.row_activate[data-v-500d50b3]:hover{background-color:#0000ff26}.lsf-space_direction_horizontal[data-v-500d50b3]{align-items:center;grid-auto-flow:column}.lsf-space_size_small[data-v-500d50b3]{grid-gap:8px}.lsf-space[data-v-500d50b3]{grid-gap:16px;display:grid}.lsf-userpic[data-v-500d50b3]{align-items:center;background:#f5f5f5;border-radius:13.5px;box-shadow:inset 0 0 0 1px rgba(0,0,0,.1);display:flex;height:28px;justify-content:center;overflow:hidden;position:relative;-webkit-user-select:none;-moz-user-select:none;user-select:none;width:28px}.lsf-userpic img[data-v-500d50b3]{font-size:12px;height:100%;line-height:22px;-o-object-fit:cover;object-fit:cover;opacity:0;position:absolute;width:100%}.lsf-annotations-list__create[data-v-500d50b3],.lsf-annotations-list__entity[data-v-500d50b3]{align-items:center;cursor:pointer;display:flex;height:48px;padding:0 16px}.lsf-userpic__username[data-v-500d50b3]{display:block;font-family:Roboto;font-size:12px;font-weight:700;line-height:22px;opacity:.4}.lsf-space_direction_vertical[data-v-500d50b3]{grid-auto-flow:row;grid-auto-rows:-webkit-max-content;grid-auto-rows:max-content;justify-content:center}.lsf-space_collapsed[data-v-500d50b3],.lsf-space_size_none[data-v-500d50b3]{grid-gap:0}.lsf-annotations-list__user[data-v-500d50b3]{display:flex;font-size:14px;line-height:16px;max-width:165px}.lsf-annotations-list__name[data-v-500d50b3]{flex:1;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.lsf-annotations-list__entity-id[data-v-500d50b3]{flex:none;font-size:11px;opacity:.5;padding-left:5px}.lsf-annotations-list__created[data-v-500d50b3]{color:rgba(0,0,0,.4);font-size:11px;line-height:13px}.mask[data-v-500d50b3]{position:fixed;width:100%;height:100%;background-color:rgba(0,0,0,.4);left:0;top:0;overflow:hidden;z-index:2000}.page-center[data-v-500d50b3]{position:absolute;width:100%;top:50%}